# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""General high level utilities for geoips processing."""

import os
from copy import deepcopy
import sys
import yaml
import logging
from glob import glob
from importlib import metadata, resources

from geoips.errors import EntryPointError, PluginError
from geoips.filenames.base_paths import PATHS as gpaths

LOG = logging.getLogger(__name__)

NAMESPACE_PREFIX = "geoips"


def get_entry_point_group(group):
    """Get entry point group."""
    if sys.version_info[:3] >= (3, 10, 0):
        return metadata.entry_points(group=group)
    else:
        return metadata.entry_points()[group]


def find_ascii_palette(name):
    """Find ASCII palette named "name".

    Search the plugins/txt/ascii_palettes directory for ASCII palettes to use
    as colormaps.
    """
    all_plugins = find_all_txt_plugins("txt/ascii_palettes")

    for plugin in all_plugins:
        if name == os.path.splitext(os.path.basename(plugin))[0]:
            return plugin
    raise ValueError(f"Non-existent txt plugin: {name}")


def find_all_txt_plugins(subdir=""):
    """Find all txt plugins in registered plugin packages.

    Search the ``plugins`` directory of each registered plugin package for files ending
    in ``.txt``. Return list of files
    """
    # Load all entry points for plugin packages
    plugin_packages = get_entry_point_group("geoips.plugin_packages")

    # Loop over the plugin packages and load all of their yaml plugins
    txt_files = []
    for pkg in plugin_packages:
        pkg_plugin_path = resources.files(pkg.value) / "plugins" / subdir
        txt_files += pkg_plugin_path.rglob("*.txt")

    return txt_files


def load_all_yaml_plugins():
    """Find all YAML plugins in registered plugin packages.

    Search the ``plugins`` directory of each registered plugin package for files ending
    in ``.yaml``. Read each plugin file
    """
    # Load all entry points for plugin packages
    plugin_packages = get_entry_point_group("geoips.plugin_packages")

    # Loop over the plugin packages and load all of their yaml plugins
    plugins = {}
    for pkg in plugin_packages:
        pkg_plugin_path = resources.files(pkg.value) / "plugins"
        yaml_files = pkg_plugin_path.rglob("*.yaml")

        # Loop over the yaml files from one package
        for yaml_file in yaml_files:
            # Load
            yaml_plugin = yaml.safe_load(open(yaml_file, "r"))

            # Set some additional information on the YAML plugin
            # The name of the package the plugin comes from
            if not yaml_plugin:
                raise PluginError(
                    f"YAML file is empty, please fill {yaml_file} with the "
                    f"appropriate information."
                )
            yaml_plugin["package"] = pkg.value
            # The relative path to the plugin within the package
            yaml_plugin["relpath"] = str(yaml_file.relative_to(pkg_plugin_path))
            # Absolute path to the plugin
            yaml_plugin["abspath"] = str(yaml_file)

            if "interface" not in yaml_plugin:
                raise PluginError(
                    f"YAML file encountered without 'interface' property: {yaml_file}"
                )
            if yaml_plugin["interface"] not in plugins:
                plugins[yaml_plugin["interface"]] = [yaml_plugin]
            else:
                plugins[yaml_plugin["interface"]].append(yaml_plugin)
    return plugins


def find_config(subpackage_name, config_basename, txt_suffix=".yaml"):
    """Find matching config file within GEOIPS packages.

    Given 'subpackage_name', 'config_basename', and txt_suffix, find matching
    text file within GEOIPS packages.

    Parameters
    ----------
    subpackage_name : str
        subdirectory under GEOIPS package to look for text file
        ie text_fname = geoips/<subpackage_name>/<config_basename><txt_suffix>
    config_basename : str
        text basename to look for,
        ie text_fname = geoips/<subpackage_name>/<config_basename><txt_suffix>
    txt_suffix : str
        suffix to look for on config file, defaults to ".yaml"
        ie text_fname = geoips/<subpackage_name>/<config_basename><txt_suffix>

    Returns
    -------
    text_fname : str
        Full path to text filename
    """
    text_fname = None

    for package_name in gpaths["GEOIPS_PACKAGES"]:
        fname = os.path.join(
            os.getenv("GEOIPS_PACKAGES_DIR"),
            package_name,
            subpackage_name,
            config_basename + txt_suffix,
        )
        # LOG.info('Trying %s', fname)
        if os.path.exists(fname):
            LOG.info("FOUND %s", fname)
            text_fname = fname
        fname = os.path.join(
            os.getenv("GEOIPS_PACKAGES_DIR"),
            package_name,
            package_name,
            subpackage_name,
            config_basename + txt_suffix,
        )
        # LOG.info('Trying %s', fname)
        if os.path.exists(fname):
            LOG.info("FOUND %s", fname)
            text_fname = fname
    return text_fname


def find_entry_point(namespace, name, default=None):
    """Find object matching 'name' using GEOIPS entry point namespace 'namespace'.

    Automatically add 'geoips' prefix to namespace for disambiguation.

    Parameters
    ----------
    namespace : str
        Entry point namespace (e.g. 'readers')
    name : str
        Entry point name (e.g. 'amsr2_netcdf')
    default : entry point, optional
        Default value if no match is found.  If this is not set (i.e. None),
        then no match will result in an exception
    """
    ep_namespace = ".".join([NAMESPACE_PREFIX, namespace])
    for ep in get_entry_point_group(ep_namespace):
        if ep.name == name:
            resolved_ep = ep.load()
            break
    else:
        resolved_ep = None
    if resolved_ep is not None:
        return resolved_ep
    else:
        if default is not None:
            return default
        else:
            raise EntryPointError(
                f"Failed to find object matching {name} in namespace {ep_namespace}"
            )


def get_all_entry_points(namespace):
    """Return all entry points in GEOIPS entry point namespace 'namespace'.

    Automatically add 'geoips' prefix to namespace for disambiguation.

    Parameters
    ----------
    namespace :str
        Entry point namespace (e.g. 'readers')
    """
    ep_namespace = ".".join([NAMESPACE_PREFIX, namespace])
    retlist = []
    # Do not use a list comprehension here so we can raise exceptions
    # containing the actual package that errored.
    try:
        for ep in get_entry_point_group(ep_namespace):
            try:
                retlist += [ep.load()]
            except Exception as resp:
                raise EntryPointError(
                    f"{resp}:"
                    f"\nAn error occurred while loading entry point "
                    f"'{ep.name}={ep.value}' entry point."
                    f"\nTry checking to ensure init file exists in package subdir "
                    f"\n(ALL directories containing python files MUST have init file)"
                ) from resp
    except KeyError:
        retlist = []
    return retlist


def list_entry_points(namespace):
    """List names of objects in GEOIPS entry point namespace 'namespace'.

    Automatically add 'geoips' prefix to namespace for disambiguation.

    Parameters
    ----------
    namespace :str
        Entry point namespace (e.g. 'readers')
    """
    ep_namespace = ".".join([NAMESPACE_PREFIX, namespace])
    return [ep.name for ep in get_entry_point_group(ep_namespace)]


def copy_standard_metadata(orig_xarray, dest_xarray, extra_attrs=None, force=True):
    """Copy standard metadata from orig_xarray to dest_xarray.

    Parameters
    ----------
    orig_xarray : xarray.Dataset
        Original xarray to copy attributes from
    dest_xarray : xarray.Dataset
        Destination xarray to copy attributes to
    extra_attrs : list of str, optional
        Additional attributes to copy, beyond the standard metadata, by default None
    force : bool, optional
        If force is True, overwrite existing attributes, by default True

    Returns
    -------
    xarray.Dataset
        dest_xarray with standard metadata copied in place from orig_xarray.
    """
    attrs = [
        "start_datetime",
        "end_datetime",
        "platform_name",
        "source_name",
        "minimum_coverage",
        "data_provider",
        "granule_minutes",
        "original_source_filenames",
        "source_file_names",
        "sample_distance_km",
        "interpolation_radius_of_influence",
        "area_definition",
    ]
    if extra_attrs is not None:
        attrs += extra_attrs

    for attr in attrs:
        if force and attr in orig_xarray.attrs.keys():
            dest_xarray.attrs[attr] = orig_xarray.attrs[attr]
        elif (
            not force
            and attr in orig_xarray.attrs.keys()
            and attr not in dest_xarray.attrs.keys()
        ):
            dest_xarray.attrs[attr] = orig_xarray.attrs[attr]


def deprecation(message):
    """Print a deprecation warning during runtime."""
    import warnings

    warnings.warn(message, DeprecationWarning, stacklevel=2)
    LOG.warning("DeprecationWarning: %s", message)


def output_process_times(process_datetimes, num_jobs=None, job_str="GeoIPS 2"):
    """Calculate and print the process times from the process_datetimes dictionary.

    Parameters
    ----------
    process_datetimes : dict
        dictionary formatted as follows:

        * ``process_datetimes['overall_start']`` - overall start datetime of
          the entire script
        * ``process_datetimes['overall_end']`` - overall end datetime of the
          entire script
        * ``process_datetimes[process_name]['start']`` - start time of an
          individual process
        * ``process_datetimes[process_name]['end']`` - end time of an
          individual process
    """
    if "overall_end" in process_datetimes and "overall_start" in process_datetimes:
        LOG.info(
            "Total Time %s: %s Num jobs: %s",
            process_datetimes["overall_end"] - process_datetimes["overall_start"],
            num_jobs,
            job_str,
        )
    for process_name in process_datetimes.keys():
        if process_name in ["overall_start", "overall_end"]:
            continue
        if "end" in process_datetimes[process_name]:
            LOG.info(
                "    SUCCESS Process Time %s: %-20s: %s",
                job_str,
                process_name,
                process_datetimes[process_name]["end"]
                - process_datetimes[process_name]["start"],
            )
        elif "fail" in process_datetimes[process_name]:
            LOG.info(
                "    FAILED  Process Time %s: %-20s: %s",
                job_str,
                process_name,
                process_datetimes[process_name]["fail"]
                - process_datetimes[process_name]["start"],
            )
        else:
            LOG.info("    MISSING Process Time %s: %s", job_str, process_name)


def replace_geoips_paths(
    fname,
    replace_paths=None,
    base_paths=None,
    curly_braces=False,
):
    """Replace standard environment variables with their non-expanded equivalents.

    Ie, replace

        * ``$HOME/geoproc/geoips_packages with $GEOIPS_PACKAGES_DIR``
        * ``$HOME/geoproc/geoips_outdirs with $GEOIPS_OUTDIRS``
        * ``$HOME/geoproc with $GEOIPS_BASEDIR``

    This allows generating output YAML fields / NetCDF attributes that can match
    between different instantiations.

    Parameters
    ----------
    fname : str
        Full path to a filename on disk
    replace_paths : list, default=None
        * Explicit list of standard variable names you would like replaced.
        * If None, replace
          ``['GEOIPS_OUTDIRS', 'GEOIPS_PACKAGES_DIR', 'GEOIPS_TESTDATA_DIR',
          'GEOIPS_DEPENDENCIES_DIR', 'GEOIPS_BASEDIR']``
    base_paths : list, default=None
        * List of PATHS dictionaries in which to find the "replace_paths" variables
        * If None, use geoips.filenames.base_paths
    curly_braces: bool, default=False
        * Specifies whether to include curly braces in the environment variables
          or not.

    Returns
    -------
    fname : str
        Path to file on disk, with explicit path replaced with environment
        variable name and/or full URL.

    Notes
    -----
    Note it replaces ALL standard variables that have a corresponding
    ``<key>_URL`` variable.

    Additionally, it replaces variables specified in "replace_paths" list with
    the unexpanded environment variable name.
    """
    # Allow multiple sets of base_path replacements
    from geoips.filenames.base_paths import PATHS as geoips_gpaths

    if base_paths is None:
        base_paths = [geoips_gpaths]

    # Replace with specified file system -> URL mapping
    for paths in base_paths:
        for key in paths.keys():
            if f"{key}_URL" in paths:
                fname = fname.replace(paths[key], paths[f"{key}_URL"])

    # Replace full paths with environment variables
    if replace_paths is None:
        replace_paths = [
            "GEOIPS_OUTDIRS",
            "GEOIPS_PACKAGES_DIR",
            "GEOIPS_TESTDATA_DIR",
            "GEOIPS_DEPENDENCIES_DIR",
            "GEOIPS_BASEDIR",
        ]
    for replace_path in replace_paths:
        for paths in base_paths:
            if replace_path in paths:
                if curly_braces:
                    fname = fname.replace(paths[replace_path], f"${{{replace_path}}}")
                else:
                    fname = fname.replace(paths[replace_path], f"${replace_path}")
    return fname


def get_required_geoips_xarray_attrs():
    """Interface deprecated v2.0."""
    required_xarray_attrs = [
        "source_name",
        "platform_name",
        "data_provider",
        "start_datetime",
        "end_datetime",
    ]
    return required_xarray_attrs


def list_product_specs_dict_yamls():
    """List all YAML files containing product params in all geoips packages.

    Returns
    -------
    list
        List of all product params dict YAMLs in all geoips packages
    """
    all_files = []
    for package_name in gpaths["GEOIPS_PACKAGES"]:
        all_files += glob(
            gpaths["GEOIPS_PACKAGES_DIR"]
            + "/"
            + package_name
            + "/*/yaml_configs/product_params/*/*.yaml"
        )
        all_files += glob(
            gpaths["GEOIPS_PACKAGES_DIR"]
            + "/"
            + package_name
            + "/yaml_configs/product_params/*/*.yaml"
        )
        all_files += glob(
            gpaths["GEOIPS_PACKAGES_DIR"]
            + "/"
            + package_name
            + "/*/yaml_configs/product_params/*.yaml"
        )
        all_files += glob(
            gpaths["GEOIPS_PACKAGES_DIR"]
            + "/"
            + package_name
            + "/yaml_configs/product_params/*.yaml"
        )
    return [fname for fname in all_files if "__init__" not in fname]


def list_product_source_dict_yamls():
    """List all YAML files containing product source specifications.

    Search in all geoips packages.

    Returns
    -------
    list
        List of all product source dict YAMLs in all geoips packages
    """
    all_files = []
    for package_name in gpaths["GEOIPS_PACKAGES"]:
        all_files += glob(
            gpaths["GEOIPS_PACKAGES_DIR"]
            + "/"
            + package_name
            + "/*/yaml_configs/product_inputs/*.yaml"
        )
        all_files += glob(
            gpaths["GEOIPS_PACKAGES_DIR"]
            + "/"
            + package_name
            + "/yaml_configs/product_inputs/*.yaml"
        )
    return [fname for fname in all_files if "__init__" not in fname]


def merge_nested_dicts(dest, src, in_place=True):
    """Perform an in-place merge of src into dest.

    Performs an in-place merge of src into dest while preserving any values that already
    exist in dest.
    """
    if not in_place:
        final_dest = deepcopy(dest)
    else:
        final_dest = dest
    try:
        final_dest.update(src | final_dest)
    except (AttributeError, TypeError):
        return
    try:
        for key, val in final_dest.items():
            try:
                merge_nested_dicts(final_dest[key], src[key])
            except KeyError:
                pass
    except AttributeError:
        raise
    if not in_place:
        return final_dest



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Procflows interface module."""

from geoips.interfaces.base import BaseModuleInterface


class ProcflowsInterface(BaseModuleInterface):
    """Class-based interface for processing workflows (procflows).

    Proclows drive a specific collection of steps for a particular type of
    processing.  Currently available are:

    * single_source: single input type and single output type.
    * overlay: two input types (one for foreground and one for background),
      with a single output type.
    * config_based: efficient method for producing all possible outputs for a
      given set of data files.
    """

    name = "procflows"
    required_args = {"standard": ["fnames"]}
    required_kwargs = {"standard": ["command_line_args"]}


procflows = ProcflowsInterface()



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Sector adjusters interface module."""

from geoips.interfaces.base import BaseModuleInterface


class SectorAdjustersInterface(BaseModuleInterface):
    """Interface for adjusting a sector based on additional information.

    Ie, pass in data and the area definition, and modify the area_definition based
    on the data itself.
    """

    name = "sector_adjusters"
    required_args = {
        "list_xarray_list_variables_to_area_def_out_fnames": [
            "xobjs",
            "area_def",
            "variables",
        ]
    }
    required_kwargs = {
        "list_xarray_list_variables_to_area_def_out_fnames": ["recenter_variables"]
    }


sector_adjusters = SectorAdjustersInterface()



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Interpolators interface module."""

from geoips.interfaces.base import BaseModuleInterface


class InterpolatorsInterface(BaseModuleInterface):
    """Interpolation routine to apply when reprojecting data."""

    name = "interpolators"
    required_args = {
        "2d": ["area_def", "input_xarray", "output_xarray", "varlist"],
        "grid": ["area_def", "input_xarray", "output_xarray", "varlist"],
    }

    required_kwargs = {"2d": ["array_num"], "grid": ["array_num"]}


interpolators = InterpolatorsInterface()



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Test script for representative product comparisons."""

import subprocess
import logging
from os.path import basename, join, splitext, dirname, isdir, isfile, exists

LOG = logging.getLogger(__name__)


def is_geotiff(fname):
    """Determine if fname is a geotiff file.

    Parameters
    ----------
    fname : str
        Name of file to check.

    Returns
    -------
    bool
        True if it is a geotiff file, False otherwise.
    """
    if splitext(fname)[-1] in [".tif"]:
        return True
    return False


def is_image(fname):
    """Determine if fname is an image file.

    Parameters
    ----------
    fname : str
        Name of file to check.

    Returns
    -------
    bool
        True if it is an image file, False otherwise.
    """
    if splitext(fname)[-1] in [".png", ".jpg", ".jpeg", ".jif"]:
        return True
    return False


def is_geoips_netcdf(fname):
    """Check if fname is a geoips formatted netcdf file.

    Parameters
    ----------
    fname : str
        Name of file to check.

    Returns
    -------
    bool
        True if it is a geoips netcdf file, False otherwise.
    """
    import xarray

    try:
        xobj = xarray.open_dataset(fname)
    except Exception:
        return False
    from geoips.geoips_utils import get_required_geoips_xarray_attrs

    return set(get_required_geoips_xarray_attrs()).issubset(set(xobj.attrs.keys()))


def is_text(fname):
    """Check if fname is a text file.

    Parameters
    ----------
    fname : str
        Name of file to check.

    Returns
    -------
    bool
        True if it is a text file, False otherwise.
    """
    if splitext(fname)[-1] in ["", ".txt", ".text", ".yaml"]:
        with open(fname) as f:
            line = f.readline()
        if isinstance(line, str):
            return True
    return False


def is_gz(fname):
    """Check if fname is a gzip file.

    Parameters
    ----------
    fname : str
        Name of file to check.

    Returns
    -------
    bool
        True if it is a gz file, False otherwise.
    """
    if splitext(fname)[-1] in [".gz"]:
        return True
    return False


def gunzip_product(fname):
    """Gunzip file fname.

    Parameters
    ----------
    fname : str
        File to gunzip.

    Returns
    -------
    str
        Filename after gunzipping
    """
    LOG.info("**** Gunzipping product for comparisons - will gzip after comparing")
    LOG.info("gunzip %s", fname)
    subprocess.call(["gunzip", fname])
    return splitext(fname)[0]


def gzip_product(fname):
    """Gzip file fname.

    Parameters
    ----------
    fname : str
        File to gzip.

    Returns
    -------
    str
        Filename after gzipping
    """
    LOG.info("**** Gzipping product - leave things as we found them")
    LOG.info("gzip %s", fname)
    subprocess.call(["gzip", fname])
    return splitext(fname)[0]


def get_out_diff_fname(compare_product, output_product, ext=None, flag=None):
    """Obtain the filename for output and comparison product diff.

    Parameters
    ----------
    compare_product : str
        Full path to product filename in the comparison directory
    output_product : str
        Full path to product filename in the current output directory
    ext : str, default=None
        Extension to use as an alternative to the original file extension
    flag : str, default=None
        Additional identifying string to include in output diff filename

    Returns
    -------
    out_diff_fname : str
        Full path to output diff file.
    """
    from os import makedirs, getenv
    from os.path import exists

    if not flag:
        flag = ""
    diffdir = join(
        dirname(compare_product), "diff_test_output_dir_{0}".format(getenv("USER"))
    )
    out_diff_fname = join(
        diffdir, "diff_test_output_" + flag + basename(output_product)
    )
    if not exists(diffdir):
        makedirs(diffdir)
    # Output jifs as tif for easy viewing.
    if ext is not None:
        out_diff_fname = splitext(out_diff_fname)[0] + ext
    elif splitext(out_diff_fname)[-1] == ".jif":
        out_diff_fname = splitext(out_diff_fname)[0] + ".png"
    return out_diff_fname


def images_match(output_product, compare_product, fuzz="5%"):
    """Use PIL and numpy to compare two images.

    Parameters
    ----------
    output_product : str
        Current output product
    compare_product : str
        Path to comparison product
    fuzz : str, optional
        NOTE: currently not implemented.
        "fuzz" argument to allow small diffs to pass - larger "fuzz" factor to make
        comparison less strict, by default 5%.

    Returns
    -------
    bool
        Return True if images match, False if they differ
    """
    exact_out_diffimg = get_out_diff_fname(
        compare_product, output_product, flag="exact_"
    )
    from PIL import Image
    import numpy as np
    from pixelmatch.contrib.PIL import pixelmatch

    LOG.info("**Comparing output_product vs. compare product")
    # Open existing images.
    out_img = Image.open(output_product)
    comp_img = Image.open(compare_product)
    diff_img = Image.new(mode="RGB", size=comp_img.size)
    # Compute the pixel diff between the two images.
    if np.array(comp_img).shape == np.array(out_img).shape:
        diff_arr = np.abs(np.array(comp_img) - np.array(out_img))
    # If shapes of arrays do not match, pixel diff can not be performed.
    # Print the names of the two images and associated shapes, and return False.
    else:
        LOG.interactive("    ***************************************")
        LOG.interactive("    *** BAD Images NOT match exactly, different sizes ***")
        LOG.interactive(
            "    ***   output_product: %s %s ***",
            np.array(out_img).shape,
            output_product,
        )
        LOG.interactive(
            "    ***   compare_product: %s %s ***",
            np.array(comp_img).shape,
            compare_product,
        )
        LOG.interactive("    ***************************************")
        return False

    # Determine the number of pixels that are mismatched
    num_pix_mismatched = pixelmatch(
        out_img, comp_img, diff_img, includeAA=True, alpha=0.33, threshold=0.05
    )
    # Currently, return 0 ONLY if the images are exactly matched.  Eventually
    # we may update this to allow returning 0 for "close enough" matches.
    if np.all(diff_arr == 0) and num_pix_mismatched == 0:
        fullimg_retval = 0
    else:
        fullimg_retval = 1
    # Write out the exact image difference.
    LOG.info("**Saving exact difference image")
    diff_img.save(exact_out_diffimg)
    LOG.info("**Done running compare")

    # If the images do not match exactly, print the output image, comparison image,
    # and exact diff image to log, for easy viewing.  Return False.
    if fullimg_retval != 0:
        LOG.interactive("    ***************************************")
        LOG.interactive("    *** BAD Images do NOT match exactly ***")
        LOG.interactive("    ***   output_product: %s ***", output_product)
        LOG.interactive("    ***   compare_product: %s ***", compare_product)
        LOG.interactive("    ***   exact dif image: %s ***", exact_out_diffimg)
        LOG.interactive("    ***************************************")
        return False

    # If the images match exactly, just output to GOOD comparison log to info level
    # (only bad comparisons to interactive level)
    if fullimg_retval != 0:
        LOG.info("    ******************************************")
        LOG.info("    *** GOOD Images match within tolerance ***")
        LOG.info("    ******************************************")
    else:
        LOG.info("    *********************************")
        LOG.info("    *** GOOD Images match exactly ***")
        LOG.info("    *********************************")

    return True


def geotiffs_match(output_product, compare_product):
    """Use diff system command to compare currently produced image to correct image.

    Parameters
    ----------
    output_product : str
        Full path to current output product
    compare_product : str
        Full path to comparison product

    Returns
    -------
    bool
        Return True if images match, False if they differ
    """
    # out_diffimg = get_out_diff_fname(compare_product, output_product)

    call_list = ["diff", output_product, compare_product]
    LOG.info("Running %s", " ".join(call_list))
    retval = subprocess.call(call_list)

    # subimg_retval = subprocess.call(call_list)
    if retval != 0:
        LOG.interactive("    *****************************************")
        LOG.interactive("    *** BAD geotiffs do NOT match exactly ***")
        LOG.interactive("    ***   output_product: %s ***", output_product)
        LOG.interactive("    ***   compare_product: %s ***", compare_product)
        LOG.interactive("    *****************************************")
        return False

    LOG.info("    ***************************")
    LOG.info("    *** GOOD geotiffs match ***")
    LOG.info("    ***************************")
    return True


def geoips_netcdf_match(output_product, compare_product):
    """Check if two geoips formatted netcdf files match.

    Parameters
    ----------
    output_product : str
        Full path to current output product
    compare_product : str
        Full path to comparison product

    Returns
    -------
    bool
        Return True if products match, False if they differ
    """
    out_difftxt = get_out_diff_fname(compare_product, output_product)
    diffout = []
    retval = True
    import xarray

    out_xobj = xarray.open_dataset(output_product)
    compare_xobj = xarray.open_dataset(compare_product)

    if out_xobj.attrs != compare_xobj.attrs:
        LOG.interactive(
            "    **************************************************************"
        )
        LOG.interactive(
            "    *** BAD GeoIPS NetCDF file attributes do NOT match exactly ***"
        )
        LOG.interactive("    ***   output_product: %s ***", output_product)
        LOG.interactive("    ***   compare_product: %s ***", compare_product)
        LOG.interactive(
            "    **************************************************************"
        )
        for attr in out_xobj.attrs.keys():
            if attr not in compare_xobj.attrs:
                diffstr = (
                    f"\nattr {attr}\n\n"
                    f"output\n{out_xobj.attrs[attr]}\n\n"
                    "not in comparison\n"
                )
                diffout += [diffstr]
                LOG.info(diffstr)
            elif out_xobj.attrs[attr] != compare_xobj.attrs[attr]:
                diffstr = (
                    f"\nattr {attr}\n\n"
                    f"output\n{out_xobj.attrs[attr]}\n\n"
                    f"comparison\n{compare_xobj.attrs[attr]}\n"
                )
                diffout += [diffstr]
                LOG.info(diffstr)
        for attr in compare_xobj.attrs.keys():
            if attr not in out_xobj.attrs:
                diffstr = (
                    f"\nattr {attr}\n\n"
                    f"not in output\n\n"
                    f"comparison\n{compare_xobj.attrs[attr]}\n"
                )
                diffout += [diffstr]
                LOG.info(diffstr)
            elif out_xobj.attrs[attr] != compare_xobj.attrs[attr]:
                diffstr = (
                    f"\nattr {attr}\n\n"
                    f"output\n{out_xobj.attrs[attr]}\n\n"
                    f"comparison\n{compare_xobj.attrs[attr]}\n"
                )
                diffout += [diffstr]
                LOG.info(diffstr)
        diffout += ["\n"]
        retval = False

    if retval is False:
        with open(out_difftxt, "w") as fobj:
            fobj.writelines(diffout)
        return False

    try:
        xarray.testing.assert_allclose(compare_xobj, out_xobj)
    except AssertionError as resp:
        LOG.interactive(
            "    ****************************************************************"
        )
        LOG.interactive(
            "    *** BAD GeoIPS NetCDF files do not match within tolerance *****"
        )
        LOG.interactive("    ***   output_product: %s ***", output_product)
        LOG.interactive("    ***   compare_product: %s ***", compare_product)
        for line in str(resp).split("\n"):
            LOG.info(f"    *** {line} ***")
        diffout += [
            "\nxarray objects do not match between current output and comparison\n"
        ]
        diffout += [f"\nOut: {out_xobj}\n"]
        diffout += [f"\nCompare: {compare_xobj}\n"]
        diffout += [f"\n{resp}\n"]
        for varname in compare_xobj.variables:
            maxdiff = (compare_xobj[varname] - out_xobj[varname]).max()
            mindiff = (compare_xobj[varname] - out_xobj[varname]).min()
            meandiff = (compare_xobj[varname] - out_xobj[varname]).mean()
            if mindiff != 0:
                LOG.info(f"    *** mindiff {varname}: {mindiff} ***")
                diffout += [f"mindiff {varname}: {mindiff}\n"]
            if maxdiff != 0:
                LOG.info(f"    *** maxdiff {varname}: {maxdiff} ***")
                diffout += [f"maxdiff {varname}: {maxdiff}\n"]
            if meandiff != 0:
                LOG.info(f"    *** meandiff {varname}: {meandiff} ***")
                diffout += [f"meandiff {varname}: {meandiff}\n"]
        LOG.info("    ****************************************************************")
        retval = False

    try:
        xarray.testing.assert_identical(compare_xobj, out_xobj)
    except AssertionError as resp:
        LOG.info("    ****************************************************************")
        LOG.info("    *** INFORMATIONAL ONLY assert_identical differences *****")
        for line in str(resp).split("\n"):
            LOG.info(f"    *** {line} ***")
        for varname in compare_xobj.variables:
            maxdiff = (compare_xobj[varname] - out_xobj[varname]).max()
            mindiff = (compare_xobj[varname] - out_xobj[varname]).min()
            meandiff = (compare_xobj[varname] - out_xobj[varname]).mean()
            if mindiff != 0:
                LOG.info(f"    *** mindiff {varname}: {mindiff} ***")
            if maxdiff != 0:
                LOG.info(f"    *** maxdiff {varname}: {maxdiff} ***")
            if meandiff != 0:
                LOG.info(f"    *** meandiff {varname}: {meandiff} ***")
        LOG.info("    ****************************************************************")

    if retval is False:
        with open(out_difftxt, "w") as fobj:
            fobj.writelines(diffout)
        return False

    return True


def text_match(output_product, compare_product):
    """Check if two text files match.

    Parameters
    ----------
    output_product : str
        Full path to current output product
    compare_product : str
        Full path to "good" comparison product

    Returns
    -------
    bool
        Return True if products match, False if they differ
    """
    ret = subprocess.run(
        ["diff", output_product, compare_product],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    LOG.debug(ret.stdout)
    if ret.returncode == 0:
        LOG.info("    *****************************")
        LOG.info("    *** GOOD Text files match ***")
        LOG.info("    *****************************")
        return True

    out_difftxt = get_out_diff_fname(compare_product, output_product)
    with open(out_difftxt, "w") as fobj:
        subprocess.call(["diff", output_product, compare_product], stdout=fobj)
    LOG.interactive("    *******************************************")
    LOG.interactive("    *** BAD Text files do NOT match exactly ***")
    LOG.interactive("    ***   output_product: %s ***", output_product)
    LOG.interactive("    ***   compare_product: %s ***", compare_product)
    LOG.interactive("    ***   out_difftxt: %s ***", out_difftxt)
    LOG.interactive("    *******************************************")
    return False


def test_product(output_product, compare_product, goodcomps, badcomps, compare_strings):
    """Test output_product against "good" product stored in "compare_path".

    Parameters
    ----------
    output_product : str
        * Full path to current output product
    compare_product : str
        * Full path to "good" comparison product
    goodcomps : list of str
        * List of full paths to all "good" successful comparisons
          (output and compare images match)
        * Each str is prepended with a "compare_string" tag to identify which
          comparison type was performed.
    badcomps : list of str
        * List of full paths to all "bad" unsuccessful comparisons
          (output and compare images differ)
        * Each str is prepended with a "compare_string" tag to identify which
          comparison type was performed.
    compare_strings : list of str
        * List of all comparison "tags" included in goodcomps and badcomps lists.
        * This list is used to remove the comparison tags from goodcomps and
          badcomps to retrieve only the file path.

    Returns
    -------
    goodcomps: list of str
        All current good comparisons appended to the list passed in.
    badcomps: list of str
        All current bad comparisons appended to the list passed in.
    compare_strings: list of str
        All current comparison "tags" added to the list passed in.

    Raises
    ------
    TypeError
        Raised when current output product does not have an associated
        comparison test defined.
    """
    matched_one = False
    if is_image(output_product):
        matched_one = True
        compare_strings += ["IMAGE "]
        if images_match(output_product, compare_product):
            goodcomps += ["IMAGE {0}".format(output_product)]
        else:
            badcomps += ["IMAGE {0}".format(output_product)]

    if is_geotiff(output_product):
        matched_one = True
        compare_strings += ["GEOTIFF "]
        if geotiffs_match(output_product, compare_product):
            goodcomps += ["GEOTIFF {0}".format(output_product)]
        else:
            badcomps += ["GEOTIFF {0}".format(output_product)]

    if is_text(output_product):
        matched_one = True
        compare_strings += ["TEXT "]
        if text_match(output_product, compare_product):
            goodcomps += ["TEXT {0}".format(output_product)]
        else:
            badcomps += ["TEXT {0}".format(output_product)]

    if is_geoips_netcdf(output_product):
        matched_one = True
        compare_strings += ["GEOIPS NETCDF "]
        if geoips_netcdf_match(output_product, compare_product):
            goodcomps += ["GEOIPS NETCDF {0}".format(output_product)]
        else:
            badcomps += ["GEOIPS NETCDF {0}".format(output_product)]

    if not matched_one:
        raise TypeError(f"MISSING TEST for output product: {output_product}")

    return goodcomps, badcomps, compare_strings


def print_gunzip_to_file(fobj, gunzip_fname):
    """Write the command to gunzip the passed "gunzip_fname" to file.

    Writes to the currently open file object, if required.
    """
    if exists(f"{gunzip_fname}.gz") and not exists(f"{gunzip_fname}"):
        fobj.write(f"gunzip -v {gunzip_fname}.gz\n")


def print_gzip_to_file(fobj, gzip_fname):
    """Write the command to gzip the passed "gzip_fname" to file.

    Writes to the currently open file object, if required.
    """
    if exists(f"{gzip_fname}.gz") and not exists(f"{gzip_fname}"):
        fobj.write(f"gzip -v {gzip_fname}\n")


def compare_outputs(compare_path, output_products, test_product_func=None):
    """Compare the "correct" imagery found the list of current output_products.

    Compares files produced in the current processing run with the list of
    "correct" files contained in "compare_path".

    Parameters
    ----------
    compare_path : str
        Path to directory of "correct" products - filenames must match output_products
    output_products : list of str
        List of strings of current output products,
        to compare with products in compare_path
    test_product_func : function, default=None
        Alternative function to be used for testing output product

          * Call signature must be:

              * output_product, compare_product, goodcomps, badcomps, compare_strings

          * Return must be:

              * goodcomps, badcomps, compare_strings

        * If None, use geoips.compare_outputs.test_product)

    Returns
    -------
    int
        Binary code: 0 if all comparisons were completed successfully.
    """
    badcomps = []
    goodcomps = []
    missingcomps = []
    missingproducts = []
    compare_strings = []
    LOG.info(
        "******************************************************************************"
    )
    LOG.info(
        "******************************************************************************"
    )
    LOG.info("*** RUNNING COMPARISONS OF KNOWN OUTPUTS IN %s ***", compare_path)
    LOG.info(
        "******************************************************************************"
    )
    LOG.info(
        "******************************************************************************"
    )
    LOG.info("")
    from glob import glob

    compare_basenames = [basename(yy) for yy in glob(compare_path + "/*")]
    final_output_products = []

    for output_product in output_products:
        if isdir(output_product):
            # Skip 'diff' output directory
            continue

        LOG.info(
            "**************************************************************************"
        )
        LOG.info("*** COMPARE  %s ***", basename(output_product))
        LOG.info(
            "**************************************************************************"
        )

        rezip = False
        if is_gz(output_product):
            rezip = True
            output_product = gunzip_product(output_product)

        if basename(output_product) in compare_basenames:
            if test_product_func is None:
                test_product_func = test_product
            goodcomps, badcomps, compare_strings = test_product_func(
                output_product,
                join(compare_path, basename(output_product)),
                goodcomps,
                badcomps,
                compare_strings,
            )
        else:
            missingcomps += [output_product]
        final_output_products += [output_product]

        # Make sure we leave things as we found them
        if rezip is True:
            gzip_product(output_product)

        LOG.info("")
    LOG.info(
        "******************************************************************************"
    )
    LOG.info(
        "******************************************************************************"
    )
    LOG.info("*** DONE RUNNING COMPARISONS OF KNOWN OUTPUTS IN %s ***", compare_path)
    LOG.info(
        "******************************************************************************"
    )
    LOG.info(
        "******************************************************************************"
    )

    product_basenames = [basename(yy) for yy in final_output_products]
    for compare_product in glob(compare_path + "/*"):
        if (
            isfile(compare_product)
            and basename(compare_product) not in product_basenames
        ):
            missingproducts += [compare_product]

    from os import makedirs, getenv
    from os.path import exists

    diffdir = join(compare_path, "diff_test_output_dir_{0}".format(getenv("USER")))
    if not exists(diffdir):
        makedirs(diffdir)

    for goodcompare in goodcomps:
        LOG.warning("GOODCOMPARE %s", goodcompare)

    for missingproduct in missingproducts:
        LOG.warning(
            "MISSINGPRODUCT no %s in output path from current run", missingproduct
        )

    for missingcompare in missingcomps:
        LOG.warning("MISSINGCOMPARE no %s in comparepath", missingcompare)

    for badcompare in badcomps:
        LOG.warning("BADCOMPARE %s", badcompare)

    if len(goodcomps) > 0:
        fname_goodcptest = join(diffdir, "cptest_GOODCOMPARE.txt")
        LOG.info("source {0}".format(fname_goodcptest))
        with open(fname_goodcptest, "w") as fobj:
            fobj.write("mkdir {0}/GOODCOMPARE\n".format(diffdir))
            for goodcomp in goodcomps:
                if compare_strings is not None:
                    for compare_string in compare_strings:
                        goodcomp = goodcomp.replace(compare_string, "")

                # For display purposes - tifs are easier to view
                out_fname = basename(goodcomp).replace(".jif", ".tif")
                print_gunzip_to_file(fobj, goodcomp)
                fobj.write(
                    "cp {0} {1}/GOODCOMPARE/{2}\n".format(goodcomp, diffdir, out_fname)
                )
                print_gzip_to_file(fobj, goodcomp)

    if len(missingcomps) > 0:
        fname_cp = join(diffdir, "cp_MISSINGCOMPARE.txt")
        fname_missingcompcptest = join(diffdir, "cptest_MISSINGCOMPARE.txt")
        LOG.interactive(
            "MISSINGCOMPARE Commands to copy %d missing files to test output path.",
            len(missingcomps),
        )
        LOG.interactive("  source {0}".format(fname_missingcompcptest))
        LOG.interactive("  source {0}".format(fname_cp))
        with open(fname_cp, "w") as fobj:
            for missingcomp in missingcomps:
                print_gunzip_to_file(fobj, missingcomp)
                fobj.write("cp -v {0} {1}/../\n".format(missingcomp, diffdir))
                print_gzip_to_file(fobj, missingcomp)
        with open(fname_missingcompcptest, "w") as fobj:
            fobj.write("mkdir {0}/MISSINGCOMPARE\n".format(diffdir))
            for missingcomp in missingcomps:
                # For display purposes - tifs are easier to view
                out_fname = basename(missingcomp).replace(".jif", ".tif")
                print_gunzip_to_file(fobj, missingcomp)
                fobj.write(
                    "cp -v {0} {1}/MISSINGCOMPARE/{2}\n".format(
                        missingcomp, diffdir, out_fname
                    )
                )
                print_gzip_to_file(fobj, missingcomp)

    if len(missingproducts) > 0:
        fname_rm = join(diffdir, "rm_MISSINGPRODUCTS.txt")
        fname_missingprodcptest = join(diffdir, "cptest_MISSINGPRODUCTS.txt")
        LOG.interactive(
            "MISSINGPRODUCTS Commands to remove %d "
            "incorrect files from test output path.",
            len(missingproducts),
        )
        LOG.interactive("  source {0}".format(fname_missingprodcptest))
        LOG.interactive("  source {0}".format(fname_rm))
        with open(fname_rm, "w") as fobj:
            for missingproduct in missingproducts:
                fobj.write("rm -v {0}\n".format(missingproduct))
        with open(fname_missingprodcptest, "w") as fobj:
            fobj.write("mkdir {0}/MISSINGPRODUCTS\n".format(diffdir))
            for missingproduct in missingproducts:
                # For display purposes - tifs are easier to view
                out_fname = basename(missingproduct).replace(".jif", ".tif")
                print_gunzip_to_file(fobj, missingproduct)
                fobj.write(
                    "cp -v {0} {1}/MISSINGPRODUCTS/{2}\n".format(
                        missingproduct, diffdir, out_fname
                    )
                )
                print_gzip_to_file(fobj, missingproduct)

    if len(badcomps) > 0:
        fname_cp = join(diffdir, "cp_BADCOMPARES.txt")
        fname_badcptest = join(diffdir, "cptest_BADCOMPARES.txt")
        LOG.interactive(
            "BADCOMPARES Commands to copy %d files that had bad comparisons",
            len(badcomps),
        )
        LOG.interactive("  source {0}".format(fname_badcptest))
        LOG.interactive("  source {0}".format(fname_cp))
        with open(fname_cp, "w") as fobj:
            for badcomp in badcomps:
                if compare_strings is not None:
                    for compare_string in compare_strings:
                        badcomp = badcomp.replace(compare_string, "")
                print_gunzip_to_file(fobj, badcomp)
                fobj.write("cp -v {0} {1}/../\n".format(badcomp, diffdir))
                print_gzip_to_file(fobj, badcomp)
        with open(fname_badcptest, "w") as fobj:
            fobj.write("mkdir {0}/BADCOMPARES\n".format(diffdir))
            for badcomp in badcomps:
                badcomp = badcomp.replace("IMAGE ", "")
                badcomp = badcomp.replace("TEXT ", "")
                badcomp = badcomp.replace("GEOIPS NETCDF ", "")
                badcomp = badcomp.replace("GEOTIFF ", "")
                # For display purposes - tifs are easier to view
                out_fname = basename(badcomp).replace(".jif", ".tif")
                print_gunzip_to_file(fobj, badcomp)
                fobj.write(
                    "cp {0} {1}/BADCOMPARES/{2}\n".format(badcomp, diffdir, out_fname)
                )
                print_gzip_to_file(fobj, badcomp)

    retval = 0
    if len(badcomps) != 0:
        curr_retval = len(badcomps)
        if len(badcomps) > 4:
            curr_retval = 4
        retval += curr_retval << 0
        LOG.info("BADCOMPS %s", len(badcomps))
    if len(missingcomps) != 0:
        curr_retval = len(missingcomps)
        if len(missingcomps) > 4:
            curr_retval = 4
        retval += curr_retval << 0
        LOG.info("MISSINGCOMPS %s", len(missingcomps))
    if len(missingproducts) != 0:
        curr_retval = len(missingproducts)
        if len(missingproducts) > 4:
            curr_retval = 4
        retval += curr_retval << 0
        LOG.info("MISSINGPRODUCTS %s", len(missingproducts))
        retval += len(missingproducts) << 4

    LOG.info("retval: %s", bin(retval))
    if retval != 0:
        LOG.info("Nonzero return value indicates error, 6 bit binary code: WXY where:")
        LOG.info(
            (
                "    Y (0-1): BADCOMP: number of bad comparisons found between"
                "comparepath and current run output path"
            )
        )
        LOG.info(
            (
                "    X (2-3): MISSINGCOMP: Number of products missing in "
                "compare path but existing in current output path"
            )
        )
        LOG.info(
            (
                "    W (4-5): MISSINGPROD: Number of products existing in "
                "comparepath, but missing in current output path"
            )
        )

    return retval



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Title formatters interface module."""

from geoips.interfaces.base import BaseModuleInterface


class TitleFormattersInterface(BaseModuleInterface):
    """Interface for creating GeoIPS formatted titles."""

    name = "title_formatters"
    required_args = {"standard": []}
    required_kwargs = {"standard": []}


title_formatters = TitleFormattersInterface()



#!/bin/env python

# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Download data from a specified URL."""

import sys
import requests


def download_test_data(url):
    """Download the specified URL and write to stdout as bytes.

    Will raise requests.exceptions.HTTPError on failure.
    """
    resp = requests.get(url, stream=True, timeout=15)
    sys.stdout.buffer.write(resp.raw.read())


if __name__ == "__main__":
    download_test_data(sys.argv[1])



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Unit tests for sector plugins."""

import pytest
from geoips import interfaces
from geoips.interfaces.yaml_based.sectors import SectorPluginBase


def test_get_all_sectors():
    """Attempt to load all available sectors."""
    interfaces.sectors.get_plugins()


@pytest.mark.parametrize("sector", interfaces.sectors.get_plugins())
def test_sector(sector):
    """Ensure that a sector has the correct attributes."""
    assert isinstance(sector, SectorPluginBase)
    assert sector.interface == "sectors"
    assert sector.name
    assert sector.family
    assert sector.__doc__
    assert sector.docstring == sector.__doc__
    # Should add this back after completing #301
    # assert isinstance(sector.area_definition, pyresample.AreaDefinition)



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Test all YAML plugins."""
import pytest
import yaml
from importlib import resources

from geoips.interfaces.base import YamlPluginValidator
from geoips.interfaces.yaml_based.products import ProductsPluginValidator
from geoips.geoips_utils import get_entry_point_group

validator = YamlPluginValidator()
product_validator = ProductsPluginValidator()


def yield_plugins():
    """Yield plugins."""
    plugin_packages = get_entry_point_group("geoips.plugin_packages")
    for pkg in plugin_packages:
        fpath = resources.files("geoips") / "plugins/yaml"
        plugin_files = fpath.rglob("*.yaml")
        for pf in plugin_files:
            yield yaml.safe_load(open(pf, "r"))


@pytest.mark.parametrize("plugin", yield_plugins())
def test_is_plugin_valid(plugin):
    """Test if plugin is valid."""
    if plugin["interface"] == "products":
        product_validator.validate(plugin)
    else:
        validator.validate(plugin)



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Geoips module for setting up logging handlers with a specified verbosity."""

import logging
import sys


def add_logging_level(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module.

    Comprehensively adds a new logging level to the `logging` module
    and the currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> add_logging_level('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        raise AttributeError("{} already defined in logging module".format(levelName))
    if hasattr(logging, methodName):
        raise AttributeError("{} already defined in logging module".format(methodName))
    if hasattr(logging.getLoggerClass(), methodName):
        raise AttributeError("{} already defined in logger class".format(methodName))

    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)


def setup_logging(logging_level="INTERACTIVE", verbose=True):
    """Set up logging handler.

    If you do this the first time with no argument, it sets up the logging
    for all submodules. Subsequently, in submodules, you can just do
    LOG = logging.getLogger(__name__)
    """
    log = logging.getLogger()
    add_logging_level("INTERACTIVE", 35)
    log.setLevel(getattr(logging, logging_level))
    fmt = logging.Formatter(
        "%(asctime)s %(module)12s.py:%(lineno)-4d %(levelname)7s: %(message)s",
        "%d_%H%M%S",
    )
    if not verbose:
        fmt = logging.Formatter("%(asctime)s: %(message)s", "%d_%H%M%S")
    stream_hndlr = logging.StreamHandler(sys.stdout)
    stream_hndlr.setFormatter(fmt)
    stream_hndlr.setLevel(logging.INFO)
    log.addHandler(stream_hndlr)
    return log



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Produce PNG test images depicting defined GeoIPS sectors."""

import os

from geoips import interfaces
from geoips.commandline.log_setup import setup_logging
from geoips.commandline.args import get_command_line_args, check_command_line_args


def main():
    """Produce PNG test images depicting defined GeoIPS sectors."""
    supported_args = ["logging_level", "sector_list", "outdir"]

    # Use standard GeoIPS argparser - use standard formatting for standard args.
    ARGS = get_command_line_args(
        arglist=supported_args,
        description="Produce PNG test images depicting defined GeoIPS sectors.",
    )

    # Get the dictionary of command line args.
    COMMAND_LINE_ARGS = ARGS.__dict__
    # Check included arguments for appropriate formatting / type.
    check_command_line_args(supported_args, COMMAND_LINE_ARGS)

    # Setup logging at the requested logging level.
    LOG = setup_logging(logging_level=COMMAND_LINE_ARGS["logging_level"])

    # For this console script, sector_list is explicitly required.
    if not COMMAND_LINE_ARGS["sector_list"]:
        raise TypeError("Must pass list of sectors")

    # Create an image for each requested sector, including just the
    # map and white background.
    for sector_name in COMMAND_LINE_ARGS["sector_list"]:
        fname = os.path.join(COMMAND_LINE_ARGS["outdir"], f"{sector_name}.png")
        sect = interfaces.sectors.get_plugin(sector_name)
        LOG.interactive(f"Creating {fname}")
        sect.create_test_plot(fname)


if __name__ == "__main__":
    main()



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Utilities for manipulating xarray Datasets and DataArrays."""

# Python Standard Libraries
import logging

# Third Party Installed Libraries

LOG = logging.getLogger(__name__)

# scipy.interpolate.griddata requires at least 4 points. Don't bother
# plotting if fewer than 4.
MIN_POINTS = 4


def get_lat_lon_points(checklat, checklon, diff, sect_xarray, varname, drop=False):
    """Pull values from xarray Datasets in specified geographic location.

    Return points a given distance around a specified lat/lon location,
    from xarray Datasets.

    Parameters
    ----------
    checklat : float
        latitude of interest
    checklon : float
        longitude of interest
    diff : float
        check +- diff of latitude and longitude
    sect_xarray : Dataset
        xarray dataset containing 'latitude' 'longitude' and varname variables
    varname : str
        variable name of data array to use for returning data values

    Returns
    -------
    float, float, int
        * min value in range
        * max value in range
        * and number of points in range
    """
    xinds = (
        (sect_xarray["longitude"] > checklon - diff)
        & (sect_xarray["longitude"] < checklon + diff)
        & (sect_xarray["latitude"] > checklat - diff)
        & (sect_xarray["latitude"] < checklat + diff)
    )
    import numpy

    if drop is True:
        retval = sect_xarray[varname].where(xinds, drop=drop).size
    else:
        retval = numpy.ma.count(sect_xarray[varname].where(xinds).to_masked_array())
    return (
        sect_xarray[varname].where(xinds).min(),
        sect_xarray[varname].where(xinds).max(),
        retval,
    )


def get_lat_lon_points_numpy(
    checklat, checklon, diff, lat_array, lon_array, data_array
):
    """Pull values from numpy arrays in specified geographic location.

    Return points a given distance around a specified lat/lon location,
    from numpy arrays.

    Parameters
    ----------
    checklat : float
        latitude of interest
    checklon : float
        longitude of interest
    diff : float
        check +- diff of latitude and longitude
    lat_array : ndarray
        numpy ndarray of latitude locations - same shape as lon_array and data_array
    lon_array : ndarray
        numpy ndarray of longitude locations - same shape as lat_array and data_array
    data_array : ndarray
        numpy ndarray data values - same shape as lat_array and lon_array

    Returns
    -------
    float, float, int
        * min value in range
        * max value in range
        * and number of points in range
    """
    import numpy

    inds = numpy.logical_and(
        lat_array > checklat - diff, lat_array < checklat + diff
    ) & numpy.logical_and(lon_array > checklon - diff, lon_array < checklon + diff)

    return (
        data_array[inds].min(),
        data_array[inds].max(),
        numpy.ma.count(data_array[inds]),
    )


def sector_xarray_temporal(
    full_xarray, mindt, maxdt, varnames, verbose=False, drop=False
):
    """Sector an xarray object temporally.  If full_xarray is None, return None.

    Parameters
    ----------
    full_xarray : xarray.Dataset
        xarray object to sector temporally
    mindt : datetime.datetime
        minimum datetime of desired data
    maxdt : datetime.datetime
        maximum datetime of desired data
    varnames : list of str
        list of variable names that should be sectored based on 'time',
        mindt, maxdt

    Returns
    -------
    xarray Dataset, or None
        * if full_xarray is None, return None
        * return full original xarray object if 'time' is not included in
          varnames list
        * else, return sectored xarray object with only the desired times,
          specified by mindt and maxdt
    """
    import numpy

    if full_xarray is None:
        LOG.info(
            "    full_xarray is None - not attempting to sector temporally, "
            "returning None"
        )
        return None

    import xarray

    time_xarray = xarray.Dataset()
    time_xarray.attrs = full_xarray.attrs.copy()

    if "time" not in varnames:
        LOG.info(
            "    time variable not included in list - not temporally sectoring, "
            "returning all data"
        )
        return full_xarray

    for varname in varnames:
        # To masked array can have a performance hit for dask arrays,
        # do not use them unnecessarily.
        # good_speeds = numpy.ma.count(full_xarray[varname].to_masked_array())
        if verbose:
            LOG.info(
                "    STARTED TIME WITH %s points for %s",
                full_xarray[varname].size,
                varname,
            )
    mindt64 = numpy.datetime64(mindt)
    maxdt64 = numpy.datetime64(maxdt)

    xarray_time_mask = (full_xarray["time"] > mindt64) & (full_xarray["time"] < maxdt64)

    time_inds = numpy.where(xarray_time_mask)

    if len(time_inds[0]) == 0:
        LOG.warning(
            "    NO TIME DATA between %s and %s for any vars, skipping", mindt, maxdt
        )
        return None

    min_y = time_inds[0].min()
    max_y = time_inds[0].max() + 1
    if len(time_inds) == 2:
        min_x = time_inds[1].min()
        max_x = time_inds[1].max() + 1

    # covg = False
    # final_good_points = 0
    for varname in varnames:
        if drop is True:
            if len(time_inds) == 2:
                # time_xarray[varname] =
                #                full_xarray[varname].where(xarray_time_mask, drop=drop)
                time_xarray[varname] = full_xarray[varname][min_y:max_y, min_x:max_x]
            else:
                # time_xarray[varname] =
                #                full_xarray[varname].where(xarray_time_mask, drop=drop)
                time_xarray[varname] = full_xarray[varname][min_y:max_y]
        else:
            time_xarray[varname] = full_xarray[varname].where(xarray_time_mask)

        # To masked array can have a performance hit for dask arrays,
        # do not use them unnecessarily.
        # good_speeds = numpy.ma.count(time_xarray[varname].to_masked_array())
        # if good_speeds > final_good_points:
        #     final_good_points = good_speeds
        # if good_speeds < MIN_POINTS:
        #     if verbose:
        #         LOG.warning('    INSUFFICIENT TIME DATA between %s and %s for var %s,
        #                     %s points', mindt, maxdt, varname, good_speeds)
        # else:
        #     if verbose:
        #         LOG.info('    MATCHED TIME %s points for var %s', good_speeds,
        #                  varname)
        #     covg = True

    # if not covg:
    if time_xarray["time"].size == 0:
        LOG.warning(
            "    INSUFFICIENT TIME DATA between %s and %s for any vars, skipping",
            mindt,
            maxdt,
        )
        return None
    LOG.info(
        "    SUFFICIENT TIME DATA between %s and %s for any var %s points",
        mindt,
        maxdt,
        time_xarray["time"].size,
    )
    return time_xarray


def sector_xarray_spatial(
    full_xarray,
    extent_lonlat,
    varnames,
    lon_pad=3,
    lat_pad=0,
    verbose=False,
    drop=False,
):
    """Sector an xarray object spatially.  If full_xarray is None, return None.

    Parameters
    ----------
    full_xarray : xarray.Dataset
        xarray object to sector spatially
    extent_lonlat : list of float
        Area to sector: [MINLON, MINLAT, MAXLON, MAXLAT]
    varnames : list of str
        list of variable names that should be sectored based on 'time'
    drop : bool
        Specify whether to remove points with no coverage (rather than masking)

    Returns
    -------
    xarray.Dataset
        * if full_xarray is None, return None,
        * else return resulting xarray Dataset.
    """
    if full_xarray is None:
        if verbose:
            LOG.info(
                "    full_xarray is None - not attempting to sector spatially, "
                "returning None"
            )
        return None

    import xarray

    sector_xarray = xarray.Dataset()
    sector_xarray.attrs = full_xarray.attrs.copy()

    import numpy
    from pyresample import utils

    if verbose:
        LOG.info("    Padding longitudes")
    # convert extent longitude to be within 0-360 range
    min_lon = utils.wrap_longitudes(extent_lonlat[0]) - lon_pad
    max_lon = utils.wrap_longitudes(extent_lonlat[2]) + lon_pad
    if min_lon > max_lon and max_lon < 0:
        max_lon = max_lon + 360
    min_lat = extent_lonlat[1] - lat_pad
    max_lat = extent_lonlat[3] + lat_pad
    if verbose:
        LOG.info("    Padding latitudes")
    # Make sure we don't extend latitudes beyond -90 / +90
    if min_lat < -90.0:
        min_lat = -90.0
    if max_lat > 90.0:
        max_lat = 90.0

    if verbose:
        LOG.info("    Wrapping longitudes")
    lons = utils.wrap_longitudes(full_xarray["longitude"])

    if verbose:
        LOG.info("    Handling dateline")
    if lons.max() > 179.5 and lons.min() < -179.5 and max_lon > 0 and min_lon > 0:
        lons = xarray.where(
            full_xarray["longitude"] < 0,
            full_xarray["longitude"] + 360,
            full_xarray["longitude"],
        )
    lats = full_xarray["latitude"]

    # To masked array can have a performance hit for dask arrays,
    # do not use them unnecessarily
    # for varname in varnames:
    #     good_speeds = numpy.ma.count(full_xarray[varname].to_masked_array())
    #     if verbose:
    #         LOG.info('    STARTED SPATIAL WITH %s points for %s', good_speeds,
    #                  varname)

    if verbose:
        LOG.info(
            "    Getting appropriate sector area lon %s to %s lat %s to %s, data "
            "minlon %s, maxlon %s, minlat %s, maxlat %s, %s points",
            min_lon,
            max_lon,
            min_lat,
            max_lat,
            lons.min().data,
            lons.max().data,
            lats.min().data,
            lats.max().data,
            lats.size,
        )
        # lons.min().data, lons.max().data, lats.min().data, lats.max().data,
        # good_speeds)

    xarray_sector_mask = (
        (lons > min_lon) & (lons < max_lon) & (lats > min_lat) & (lats < max_lat)
    )

    sector_inds = numpy.where(xarray_sector_mask)

    if not len(sector_inds[0]):
        LOG.warning(
            "    NO SPATIAL DATA between %0.2f and %0.2f lon and %0.2f and %0.2f lat",
            min_lon,
            max_lon,
            min_lat,
            max_lat,
        )
        return None

    min_y = sector_inds[0].min()
    max_y = sector_inds[0].max() + 1
    if len(sector_inds) == 2:
        min_x = sector_inds[1].min()
        max_x = sector_inds[1].max() + 1

    # covg = False
    # final_good_points = 0
    for varname in varnames:
        if drop is True:
            if len(sector_inds) == 2:
                # sector_xarray[varname] =
                #              full_xarray[varname].where(xarray_sector_mask, drop=drop)
                sector_xarray[varname] = full_xarray[varname][min_y:max_y, min_x:max_x]
            else:
                # sector_xarray[varname] =
                #              full_xarray[varname].where(xarray_sector_mask, drop=drop)
                sector_xarray[varname] = full_xarray[varname][min_y:max_y]
        else:
            sector_xarray[varname] = full_xarray[varname].where(xarray_sector_mask)

        # To masked array can have a performance hit for dask arrays,
        # do not use them unnecessarily
        # good_speeds = numpy.ma.count(sector_xarray[varname].to_masked_array())
        # if good_speeds > final_good_points:
        #     final_good_points = good_speeds

        # if sector_xarray[varname].size < MIN_POINTS or good_speeds < MIN_POINTS:
        #     if verbose:
        #         LOG.warning('    INSUFFICIENT SPATIAL DATA between %0.2f and %0.2f lon
        #                     and %0.2f and %0.2f lat, varname %s, %s points',
        #                     extent_lonlat[0], extent_lonlat[2], extent_lonlat[1],
        #                     extent_lonlat[3], varname, good_speeds)
        # else:
        #     if verbose:
        #         LOG.info('    MATCHED SPATIAL %s points for var %s after location
        #                  sectoring', good_speeds, varname)
        #     covg = True

    # if covg:
    if sector_xarray["latitude"].size == 0:
        LOG.warning(
            "    OVERALL INSUFFICIENT SPATIAL DATA between %0.2f and %0.2f lon"
            " and %0.2f and %0.2f lat",
            min_lat,
            max_lon,
            min_lat,
            max_lat,
        )
        return None

    LOG.info(
        "    OVERALL SUFFICIENT SPATIAL DATA between %0.2f and %0.2f lon and %0.2f"
        " and %0.2f lat %0.2f points",
        min_lon,
        max_lon,
        min_lat,
        max_lat,
        sector_xarray["latitude"].size,
    )
    # extent_lonlat[0], extent_lonlat[2], extent_lonlat[1], extent_lonlat[3],
    # final_good_points)
    return sector_xarray


def sector_xarray_dataset(
    full_xarray,
    area_def,
    varnames,
    lon_pad=3,
    lat_pad=0,
    verbose=False,
    hours_before_sector_time=18,
    hours_after_sector_time=6,
    drop=False,
    window_start_time=None,
    window_end_time=None,
):
    """Use the xarray to appropriately sector out data by lat/lon and time.

    Parameters
    ----------
    full_xarray: xarray.Dataset
        The full input xarray Dataset
    area_def: pyresample.AreaDefinition
        The requested region to sector spatially from the full_xarray.
        If None, no spatial sectoring required.
    varnames: list of str
        List of variable names within full_xarray to include in the final
        xarray Dataset
    lon_pad: float, default=3
        Amount of padding to include around the longitude dimension
    lat_pad: float, default=0
        Amount of padding to include around the latitude dimension
    verbose: bool, default=False
        If True, include log output, otherwise return silently.
    hours_before_sector_time: float, default=18
        For dynamic sectors, number of hours before sector start time
        to include when sectoring temporally.
    hours_after_sector_time: float, default=6
        For dynamic sectors, number of hours after sector end time
        to include when sectoring temporally.
    drop: bool, default=False
        If True, drop data outside range, temporally and spatially.
        If False, mask outside range.
    window_start_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.
        hours_before_sector_time and hours_after_sector_time are ignored if
        window start/end time are set!
    window_end_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.
        hours_before_sector_time and hours_after_sector_time are ignored if
        window start/end time are set!

    Returns
    -------
    xarray.Dataset
        sectored dataset containing requested "varnames" sectored spatially and
        temporally as requested.
    """
    from datetime import timedelta

    LOG.debug(
        "Full xarray start/end datetime: %s %s",
        full_xarray.start_datetime,
        full_xarray.end_datetime,
    )
    # numpy.ma.count(full_xarray[varnames[0]].to_masked_array()))

    if area_def is not None:
        if (
            hasattr(area_def, "sector_start_datetime")
            and area_def.sector_start_datetime
        ):
            if not window_start_time:
                # If it is a dynamic sector and we did not specify window
                # start/end time, sector temporally to make sure we use the
                # appropriate data
                mindt = area_def.sector_start_datetime - timedelta(
                    hours=hours_before_sector_time
                )
                maxdt = area_def.sector_start_datetime + timedelta(
                    hours=hours_after_sector_time
                )
            else:
                # If we requested window start/end time, use it.
                mindt = window_start_time
                maxdt = window_end_time
            time_xarray = sector_xarray_temporal(
                full_xarray, mindt, maxdt, varnames, verbose=verbose
            )
        elif window_start_time:
            # If this is not a dynamic sector, but we requested start/end
            # time, still apply the requested window.
            mindt = window_start_time
            maxdt = window_end_time
            time_xarray = sector_xarray_temporal(
                full_xarray, mindt, maxdt, varnames, verbose=verbose
            )
        else:
            # If it is not a dynamic sector, and we didn't pass window
            # start/end time, just return all of the data, because all
            # we care about is spatial coverage.
            time_xarray = full_xarray.copy()

        extent_lonlat = list(area_def.area_extent_ll)
        sector_xarray = sector_xarray_spatial(
            time_xarray,
            extent_lonlat,
            varnames,
            lon_pad,
            lat_pad,
            verbose=verbose,
            drop=drop,
        )
        if (
            sector_xarray is not None
            and "time" in varnames
            and hasattr(area_def, "sector_start_datetime")
            and area_def.sector_start_datetime
        ):
            from geoips.xarray_utils.time import (
                get_min_from_xarray_time,
                get_max_from_xarray_time,
            )

            sector_xarray.attrs["area_definition"] = area_def
            sector_xarray.attrs["start_datetime"] = get_min_from_xarray_time(
                sector_xarray, "time"
            )
            sector_xarray.attrs["end_datetime"] = get_max_from_xarray_time(
                sector_xarray, "time"
            )
        elif sector_xarray is not None:
            sector_xarray.attrs["area_definition"] = area_def
            sector_xarray.attrs["start_datetime"] = full_xarray.start_datetime
            sector_xarray.attrs["end_datetime"] = full_xarray.end_datetime

    elif window_start_time:
        # If the area_def is not specified, but we requested start and end time,
        # apply requested window here.
        mindt = window_start_time
        maxdt = window_end_time
        sector_xarray = sector_xarray_temporal(
            full_xarray, mindt, maxdt, varnames, verbose=verbose
        )
    else:
        # If we did not request temporal or spatial sectoring,
        # just return full xarray
        sector_xarray = full_xarray.copy()

    if sector_xarray is not None:
        if verbose:
            LOG.info(
                "Sectored data start/end datetime: %s %s",
                sector_xarray.start_datetime,
                sector_xarray.end_datetime,
            )
            # numpy.ma.count(full_xarray[varnames[0]].to_masked_array()))

    return sector_xarray


def get_vis_ir_bg(sect_xarray):
    """Find matching vis/ir background for data in sect_xarray."""
    from geoips.data_manipulations.merge import get_matching_files
    from geoips.filenames.base_paths import PATHS as gpaths

    irfnames = get_matching_files(
        sect_xarray.area_definition.area_id,
        [sect_xarray.area_definition.area_id],
        ["himawari-8", "goes-16", "goes-17", "msg-1", "msg-4"],
        ["ahi", "abi", "abi", "seviri", "seviri"],
        [30, 30, 30, 30, 30],
        gpaths["PRECALCULATED_DATA_PATH"],
        sect_xarray.start_datetime,
        "Infrared-Gray_latitude_longitude",
        time_format="%H%M*",
        buffer_mins=75,
        verbose=False,
        single_match=True,
    )
    visfnames = get_matching_files(
        sect_xarray.area_definition.area_id,
        [sect_xarray.area_definition.area_id],
        ["himawari-8", "goes-16", "goes-17", "msg-1", "msg-4"],
        ["ahi", "abi", "abi", "seviri", "seviri"],
        [30, 30, 30, 30, 30],
        gpaths["PRECALCULATED_DATA_PATH"],
        sect_xarray.start_datetime,
        "Visible_latitude_longitude",
        time_format="%H%M*",
        buffer_mins=75,
        verbose=False,
        single_match=True,
    )
    LOG.info("irfnames: %s", irfnames)
    LOG.info("visfnames: %s", visfnames)
    ret_arr = {}
    from geoips.plugins.modules.readers.geoips_netcdf import read_xarray_netcdf

    if irfnames:
        ir_bg = read_xarray_netcdf(irfnames[0])
        ret_arr["IR"] = ir_bg
        ret_arr["METADATA"] = ir_bg[[]]
    if visfnames:
        vis_bg = read_xarray_netcdf(visfnames[0])
        ret_arr["VIS"] = vis_bg
        ret_arr["METADATA"] = vis_bg[[]]
    LOG.info("ret_arr: %s", ret_arr)
    return ret_arr


def sector_xarrays(
    xobjs,
    area_def,
    varlist,
    verbose=False,
    hours_before_sector_time=18,
    hours_after_sector_time=6,
    check_center=True,
    drop=False,
    lon_pad=3,
    lat_pad=0,
    window_start_time=None,
    window_end_time=None,
):
    """Return dict of sectored xarray objects.

    Parameters
    ----------
    xobjs: dict of xarray.Dataset
        The full input xarray Datasets, with dataset names as keys to the dictionary.
    area_def: pyresample.AreaDefinition
        The requested region to sector spatially from the full_xarray.
        If None, no spatial sectoring required.
    varlist: list of str
        List of variable names within xarray Datasets to include in the final
        sectored xarray Datasets
    verbose: bool, default=False
        If True, include log output, otherwise return silently.
    hours_before_sector_time: float, default=18
        For dynamic sectors, number of hours before sector start time
        to include when sectoring temporally.
    hours_after_sector_time: float, default=6
        For dynamic sectors, number of hours after sector end time
        to include when sectoring temporally.
    check_center: bool, default=True
        If True and sector_type is "tc", check that there is coverage in the
        center of the image prior to sectoring. Skip dataset altogether if no
        center coverage.
    drop: bool, default=False
        If True, drop data outside range, temporally and spatially.
        If False, mask outside range.
    lon_pad: float, default=3
        Amount of padding to include around the longitude dimension
    lat_pad: float, default=0
        Amount of padding to include around the latitude dimension
    window_start_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.
        hours_before_sector_time and hours_after_sector_time are ignored if
        window start/end time are set!
    window_end_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.
        hours_before_sector_time and hours_after_sector_time are ignored if
        window start/end time are set!

    Returns
    -------
    dict of xarray.Dataset
        Dictionary of sectored datasets containing requested "varnames",
        sectored spatially and temporally as requested.
    """
    ret_xobjs = {}
    for key, xobj in xobjs.items():
        LOG.info("SECTORING dataset %s area_def %s", key, area_def.name)
        LOG.info(" requested variables %s", set(varlist))
        LOG.info(" dataset variables %s", set(xobj.variables.keys()))
        LOG.info(" dataset data_vars %s", set(xobj.data_vars))
        # Compile a list of variables that will be used to sector - the current data
        # variable, and we will add in the appropriate latitude and longitude variables
        # (of the same shape as data), and if it exists the appropriately shaped time
        # array.
        # Use data_vars here - since coordinates (lat/lon/time) get added below
        vars_to_interp = list(set(varlist) & set(xobj.data_vars))
        if not vars_to_interp:
            LOG.info("  SKIPPING no required variables in dataset %s", key)
            continue

        from geoips.sector_utils.utils import is_dynamic_sector

        if is_dynamic_sector(area_def):
            LOG.info(
                "  Trying to sector %s with dynamic time %s, data time %s to %s, "
                "%s points",
                area_def.area_id,
                area_def.sector_start_datetime,
                xobj.start_datetime,
                xobj.end_datetime,
                xobj["latitude"].size,
            )
        else:
            LOG.info(
                "  Trying to sector %s, %s points",
                area_def.area_id,
                xobj["latitude"].size,
            )

        vars_to_sect = []
        vars_to_sect += vars_to_interp
        # we have to have 'latitude','longitude" in the full_xarray, and 'time' if we
        # want temporal sectoring.
        # Note if lat/lon/time are included as coordinates, they will NOT show up
        # in data_vars, so must use variables
        if "latitude" in list(xobj.variables.keys()):
            vars_to_sect += ["latitude"]
        if "longitude" in list(xobj.variables.keys()):
            vars_to_sect += ["longitude"]
        if "time" in list(xobj.variables.keys()):
            vars_to_sect += ["time"]

        from geoips.xarray_utils.data import sector_xarray_dataset

        # The list of variables in vars_to_sect must ALL be the same shape
        sect_xarray = sector_xarray_dataset(
            xobj,
            area_def,
            vars_to_sect,
            verbose=verbose,
            hours_before_sector_time=hours_before_sector_time,
            hours_after_sector_time=hours_after_sector_time,
            drop=drop,
            lon_pad=lon_pad,
            lat_pad=lat_pad,
            window_start_time=window_start_time,
            window_end_time=window_end_time,
        )

        # numpy arrays fail if numpy_array is None, and xarrays fail if x_array == None
        if sect_xarray is None:
            if verbose:
                LOG.info("  No coverage - skipping dataset %s", key)
            continue

        # check for any obs  near TC center (with a box of 8deg X 8deg), i.e.,
        # within a  range of 400km)
        from geoips.sector_utils.utils import is_sector_type

        if is_sector_type(area_def, "tc") and check_center:
            from geoips.sector_utils.utils import check_center_coverage

            has_covg, covg_xarray = check_center_coverage(
                sect_xarray,
                area_def,
                varlist=vars_to_sect,
                covg_varlist=vars_to_sect,
                width_degrees=8,
                height_degrees=8,
                verbose=verbose,
            )
            if not has_covg:
                LOG.info(
                    "  SKIPPING DATASET %s, NO COVERAGE IN center box - NOT PROCESSING",
                    key,
                )
                continue

            # If the time within the box is > 50 min, we have two overpasses. ALL PMW
            # sensors are polar orbiters.
            if (covg_xarray.end_datetime - covg_xarray.start_datetime).seconds > 3000:
                LOG.info(
                    "Original sectored xarray contains more than one overpass - "
                    "switching to start/datetime in center"
                )
                sect_xarray.attrs["start_datetime"] = covg_xarray.start_datetime
                sect_xarray.attrs["end_datetime"] = covg_xarray.end_datetime

        sect_xarray.attrs[
            "area_definition"
        ] = area_def  # add name of this sector to sector attribute
        if hasattr(sect_xarray, "time"):
            from geoips.xarray_utils.time import get_min_from_xarray_time
            from geoips.xarray_utils.time import get_max_from_xarray_time

            sect_xarray.attrs["start_datetime"] = get_min_from_xarray_time(
                sect_xarray, "time"
            )
            sect_xarray.attrs["end_datetime"] = get_max_from_xarray_time(
                sect_xarray, "time"
            )
            # Note:  need to test whether above two lines can reselect min and max
            # time_info for this sector

        LOG.debug(
            "  Sectored data start/end datetime: %s %s, %s points from var %s, "
            "all vars %s",
            sect_xarray.start_datetime,
            sect_xarray.end_datetime,
            sect_xarray[vars_to_interp[0]].size,
            vars_to_interp[0],
            vars_to_interp,
        )
        sect_xarray.attrs["sectored"] = True
        ret_xobjs[key] = sect_xarray
        ret_xobjs["METADATA"] = sect_xarray[[]]

    return ret_xobjs


def get_sectored_xarrays(
    xobjs, area_def, varlist, get_bg_xarrays=False, check_center=True, drop=False
):
    """Get all xarray objects sectored to area_def.

    Return primary dataset, as well as VIS/IR overlay datasets.
    """
    # All datasets will be of the same data type
    sect_xarrays = sector_xarrays(
        xobjs, area_def, varlist, check_center=check_center, drop=drop
    )
    if get_bg_xarrays and sect_xarrays:
        bg_xarrays = get_vis_ir_bg(sect_xarrays["METADATA"])
        sect_xarrays += sector_xarrays(
            bg_xarrays,
            area_def,
            ["Infrared-Gray", "Visible"],
            check_center=check_center,
            drop=drop,
        )
    else:
        LOG.info(
            "SKIPPING BACKGROUNDS, no coverage for %s %s",
            xobjs[0].source_name,
            area_def.name,
        )
    return sect_xarrays



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""geoips.plugins.modules init file."""



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Command line script for updating the TC database."""


def main():
    """Update TC tracks database via command line."""
    import sys
    from geoips.commandline.log_setup import setup_logging

    LOG = setup_logging()
    from geoips.sector_utils.tc_tracks_database import check_db

    if len(sys.argv) > 2:
        LOG.info("Calling with arg %s", sys.argv[1:])
        check_db(sys.argv[1:])
    else:
        check_db()


if __name__ == "__main__":
    main()



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Utilities for tracking and monitoring memory and resource usage."""
import logging
import socket

import resource

LOG = logging.getLogger(__name__)


try:
    import psutil
except ImportError:
    LOG.info(
        "Failed import psutil in utils/memusg.py. " + "If you need it, install it."
    )


def print_mem_usage(logstr="", verbose=False):
    """Print memory usage to LOG.info.

    * By default include psutil output.
    * If verbose is True, include output from both psutil and resource packages.
    """
    # If psutil / socket / resource are not imported, do not fail
    usage_dict = {}
    try:
        vmem_percent = psutil.virtual_memory().percent
        LOG.info(
            "virtual perc: %s on %s %s",
            str(vmem_percent),
            str(socket.gethostname()),
            logstr,
        )
        swap_percent = psutil.swap_memory().percent
        LOG.info(
            "swap perc:    %s on %s %s",
            str(swap_percent),
            str(socket.gethostname()),
            logstr,
        )
    except NameError as resp:
        vmem_percent = swap_percent = "nan"
        LOG.info(
            "%s: psutil or socket not defined, no percent memusg output", str(resp)
        )
    try:
        highest = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        LOG.info(
            "highest:      %s on %s %s",
            str(highest),
            str(socket.gethostname()),
            logstr,
        )
    except NameError as resp:
        highest = "nan"
        LOG.info(
            "%s: resource or socket not defined, no highest memusg output", str(resp)
        )

    if verbose:
        usage_dict = print_resource_usage(logstr)
    usage_dict["memusg_virtual"] = vmem_percent
    usage_dict["memusg_swap"] = swap_percent
    usage_dict["memusg_highest"] = highest
    return usage_dict


def print_resource_usage(logstr=""):
    """Print verbose resource usage, using "resource" package."""
    usage_dict = {}
    try:
        usage = resource.getrusage(resource.RUSAGE_SELF)
        for name, desc in [
            ("ru_utime", "RESOURCE " + logstr + " User time"),
            ("ru_stime", "RESOURCE " + logstr + " System time"),
            ("ru_maxrss", "RESOURCE " + logstr + " Max. Resident Set Size"),
            ("ru_ixrss", "RESOURCE " + logstr + " Shared Memory Size"),
            ("ru_idrss", "RESOURCE " + logstr + " Unshared Memory Size"),
            ("ru_isrss", "RESOURCE " + logstr + " Stack Size"),
            ("ru_inblock", "RESOURCE " + logstr + " Block inputs"),
            ("ru_oublock", "RESOURCE " + logstr + " Block outputs"),
        ]:
            LOG.info("%-25s (%-10s) = %s", desc, name, getattr(usage, name))
            usage_dict[name] = getattr(usage, name)
    except NameError:
        LOG.info("resource not defined")
    return usage_dict



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Algorithms interface module."""

from geoips.interfaces.base import BaseModuleInterface


class AlgorithmsInterface(BaseModuleInterface):
    """Data manipulations to apply to the dataset."""

    name = "algorithms"

    required_args = {
        "scalar_to_scalar": [],
        "single_channel": ["arrays"],
        "channel_combination": ["arrays"],
        "list_numpy_to_numpy": ["arrays"],
        "xarray_to_numpy": ["xobj"],
        "xarray_to_xarray": ["xobj", "variables", "product_name"],
        "rgb": ["arrays"],
        "xarray_dict_to_xarray": ["xarray_dict"],
        "xarray_dict_dict_to_xarray": ["xarray_dict_dict"],
        "xarray_dict_to_xarray_dict": ["xarray_dict"],
        "xarray_dict_area_def_to_numpy": ["xarray_dict", "area_def"],
    }

    required_kwargs = {
        "scalar_to_scalar": ["value"],
        "single_channel": [],
        "channel_combination": [],
        "xarray_to_numpy": [],
        "xarray_to_xarray": [],
        "list_numpy_to_numpy": [],
        "rgb": [],
        "xarray_dict_to_xarray": [],
        "xarray_dict_dict_to_xarray": [],
        "xarray_dict_to_xarray_dict": [],
        "xarray_dict_area_def_to_numpy": [],
    }


algorithms = AlgorithmsInterface()



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Sector spec generators interface module."""

from geoips.interfaces.base import BaseModuleInterface


class SectorSpecGeneratorsInterface(BaseModuleInterface):
    """Interface for generating appropriate specifications for a sector.

    Generates a sector specification (location, resolution, etc)
    using a dynamically determined subset of information (ie, center
    lat/lon of the region determined at run time, while the resolution
    and width/height are predetermined)
    """

    name = "sector_spec_generators"
    required_args = {"area_definition": []}
    required_kwargs = {"area_definition": []}


sector_spec_generators = SectorSpecGeneratorsInterface()



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""GeoIPS interface module."""

from geoips.interfaces.module_based.algorithms import algorithms
from geoips.interfaces.module_based.colormappers import colormappers
from geoips.interfaces.module_based.coverage_checkers import coverage_checkers
from geoips.interfaces.module_based.filename_formatters import filename_formatters
from geoips.interfaces.module_based.interpolators import interpolators
from geoips.interfaces.module_based.output_formatters import output_formatters
from geoips.interfaces.module_based.procflows import procflows
from geoips.interfaces.module_based.readers import readers
from geoips.interfaces.module_based.sector_adjusters import sector_adjusters
from geoips.interfaces.module_based.sector_metadata_generators import (
    sector_metadata_generators,
)
from geoips.interfaces.module_based.sector_spec_generators import sector_spec_generators
from geoips.interfaces.module_based.title_formatters import title_formatters

from geoips.interfaces.yaml_based.feature_annotators import feature_annotators
from geoips.interfaces.yaml_based.gridline_annotators import gridline_annotators
from geoips.interfaces.yaml_based.product_defaults import product_defaults
from geoips.interfaces.yaml_based.products import products
from geoips.interfaces.yaml_based.sectors import sectors

__all__ = [
    "algorithms",
    "colormappers",
    "coverage_checkers",
    "feature_annotators",
    "filename_formatters",
    "gridline_annotators",
    "interpolators",
    "output_formatters",
    "procflows",
    "product_defaults",
    "products",
    "readers",
    "sector_adjusters",
    "sector_metadata_generators",
    "sector_spec_generators",
    "sectors",
    "title_formatters",
]



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Utils to handle time stamp information within xarray objects."""


def get_posix_from_datetime(dt):
    """Return the POSIX timestamp in seconds.

    Parameters
    ----------
    dt : datetime.datetime
        datetime object to convert to posix timestamp

    Returns
    -------
    long
        representing seconds since 1 January 1970 at 00Z (epoch seconds)
    """
    import calendar

    return calendar.timegm(dt.timetuple())


def get_datetime_from_datetime64(dt64):
    """Get a python datetime object from a numpy datetime64 object.

    Parameters
    ----------
    dt64 : numpy.datetime64'
        numpy.datetime64 object

    Returns
    -------
    datetime.datetime
        Python datetime object

    Notes
    -----
    Backwards compatible with numpy versions
    """
    scale = 1e-9
    if "[ns]" in dt64.dtype.name:
        scale = 1e-9
    from datetime import datetime

    return datetime.utcfromtimestamp(dt64.astype(int) * scale)


def get_min_from_xarray_time(xarray_obj, varname):
    """Get the minimum time as a datetime object from xarray object.

    Parameters
    ----------
    xarray_obj : xarray.Dataset or xarray.DataArray
        xarray object from which to extract the minimum time
    varname : str
        Timestamp variable name from which to extract the minimum time

    Returns
    -------
    datetime.datetime
        Python datetime.datetime object representing minimum time of the
        Dataset or DataArray
    """
    minval = xarray_obj[varname].min().to_dict()["data"]
    # Hack to get around bug in most recent version of numpy
    if minval is None:
        import numpy

        goodinds = numpy.ma.where(xarray_obj[varname].to_masked_array())
        minval = get_datetime_from_datetime64(
            xarray_obj[varname].to_masked_array()[goodinds].min()
        )
    return minval


def get_max_from_xarray_time(xarray_obj, varname):
    """Get the maximum time as a datetime object from xarray object.

    Parameters
    ----------
    xarray_obj : xarray.Dataset or xarray.DataArray
        xarray object from which to extract the maximum time
    varname : str
        Timestamp variable name from which to extract the maximum time

    Returns
    -------
    datetime.datetime
        Python datetime.datetime object representing maximum time of the
        Dataset or DataArray
    """
    maxval = xarray_obj[varname].max().to_dict()["data"]
    # Hack to get around bug in most recent version of numpy
    if maxval is None:
        import numpy

        goodinds = numpy.ma.where(xarray_obj[varname].to_masked_array())
        maxval = get_datetime_from_datetime64(
            xarray_obj[varname].to_masked_array()[goodinds].max()
        )
    return maxval



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Colormappers interface module."""

from geoips.interfaces.base import BaseModuleInterface


class ColormappersInterface(BaseModuleInterface):
    """Interface for the colormapper to apply to the resulting product."""

    name = "colormappers"
    required_args = {
        "matplotlib": [],
    }

    required_kwargs = {
        "matplotlib": [],
    }

    allowable_kwargs = {
        "matplotlib": [
            "data_range",
            "cmap_name",
            "ascii_path",
            "create_colorbar",
            "cbar_label",
            "cbar_ticks",
            "cbar_tick_labels",
            "cbar_spacing",
            "cbar_full_width",
            "colorbar_kwargs",
            "set_ticks_kwargs",
            "set_label_kwargs",
        ]
    }


colormappers = ColormappersInterface()



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""
The Geolocated Information Processing System (GeoIPS).

GeoIPS |unireg| Base Package
----------------------------

The GeoIPS Base Package provides a Python 3 based architecture supporting a wide
variety of satellite and weather data processing. The modular nature of the GeoIPS
base infrastructure also allows plug-and-play capability for user-specified custom
functionality.

Homepage: https://github.com/NRLMMD-GEOIPS/geoips

.. |unireg|    unicode:: U+000AE .. REGISTERED SIGN
"""

from geoips import errors
from geoips import filenames
from geoips import interfaces
from geoips import utils
from geoips import xarray_utils
from ._version import __version__, __version_tuple__

__all__ = [
    "interfaces",
    "errors",
    "filenames",
    "utils",
    "xarray_utils",
    "__version__",
    "__version_tuple__",
]
